# Copyright (c) 2000, 2001 Phil Thompson <phil@river-bank.demon.co.uk>
#
# This file is part of BlackAdder and subject to the same license conditions.


import sys


# Check the Python version is at least 2.0.

try:
    if sys.hexversion < 0x2000000:
        sys.exit(8)
except:
    sys.exit(8)


import socket
import select
import codeop
import traceback
import types
import bdb
import os


# PyQt may not be installed (eg. this may be the demo version).

try:
    from qt import SIGNAL, PYSIGNAL, QObject, QSocketNotifier, qApp
except:
    sys.exit(9)

# The protocol "words".
RequestLoad = '>Load<'
RequestContinue = '>Continue<'
RequestStep = '>Step<'
RequestBreak = '>Break<'
RequestVariables = '>Variables<'
RequestStepOver = '>StepOver<'


ResponseException = '>Exception<'
ResponseSyntax = '>Syntax<'
ResponseExit = '>Exit<'
ResponseLine = '>Line<'
ResponseSysPath = '>SysPath<'
ResponseVariables = RequestVariables


DebugClientInstance = None

def DebugClientPreEventLoopHook():
    """Called by PyQt just before the Qt event loop is entered.
    """
    if DebugClientInstance is not None:
        DebugClientInstance.preEventLoopHook()
 
def DebugClientPostEventLoopHook():
    """Called by PyQt just after the Qt event loop is left.
    """
    if DebugClientInstance is not None:
        DebugClientInstance.postEventLoopHook()
 
# Make the hooks available to PyQt.
__builtins__.__dict__['__pyQtPreEventLoopHook__'] = DebugClientPreEventLoopHook
__builtins__.__dict__['__pyQtPostEventLoopHook__'] = DebugClientPostEventLoopHook
 

def DebugClientRawInput(prompt):
    """A replacement for the standard raw_input builtin that works with the
    event loop.
    """
    if DebugClientInstance is None:
        return DebugClientOrigRawInput(prompt)
    return DebugClientInstance.raw_input(prompt)

# Use our own raw_input().
DebugClientOrigRawInput = __builtins__.__dict__['raw_input']
__builtins__.__dict__['raw_input'] = DebugClientRawInput
 

class AsyncFile: # {{{1
    """AsyncFile(self,sock,mode,name)
 
    A class that wraps a socket object with a file object interface.  sock is
    the socket object being wrapped.  mode is the value to set the mode
    attribute to.  name is the value to set the name attribute to.
 
    """
    def __init__(self,sock,mode,name):
        # Initialise the attributes.
        self.closed = 0
        self.sock = sock
        self.mode = mode
        self.name = name
        self.softspace = 0
 
        self.wpending = ''
 
    def checkMode(self,mode):
        """
        Private method to check that an operation is permitted according to the
        mode of the file.
        """
        if mode != self.mode:
            raise IOError, '[Errno 9] Bad file descriptor'
 
    def nWrite(self,n):
        """
        Private method to write a specific number of pending bytes.
        """
        if n:
            sent = self.sock.send(self.wpending[:n])
            self.wpending = self.wpending[sent:]
 
    def pendingWrite(self):
        """
        Public method that returns the number of bytes waiting to be written.
        """
        return self.wpending.rfind('\n') + 1
 
    def close(self):
        if not self.closed:
            self.flush()
            self.sock.close()
            self.closed = 1
 
    def flush(self):
        self.nWrite(len(self.wpending))
 
    def isatty(self):
        return 0
 
    def fileno(self):
        return self.sock.fileno()
 
    def read(self,size=-1):
        self.checkMode('r')
 
        if size < 0:
            size = 20000
 
        return self.sock.recv(size)
 
    def readline(self,size=-1):
        # Note that this method will not block and may return only a part of a
        # line if that is all that is available.
        self.checkMode('r')
 
        if size < 0:
            size = 20000
 
        return self.sock.recv(size)
 
    def readlines(self,sizehint=-1):
        lines = []
        room = sizehint
 
        line = self.readline(room)
        linelen = len(line)
 
        while linelen > 0:
            lines.append(line)
 
            if sizehint >= 0:
                room = room - linelen
 
                if room <= 0:
                    break
 
            line = self.readline(room)
            linelen = len(line)
 
        return lines
 
    def seek(self,offset,whence=0):
        raise IOError, '[Errno 29] Illegal seek'
 
    def tell(self):
        raise IOError, '[Errno 29] Illegal seek'
 
    def truncate(self,size=-1):
        raise IOError, '[Errno 29] Illegal seek'
 
    def write(self,str):
        self.checkMode('w')
        self.wpending = self.wpending + str
        self.nWrite(self.pendingWrite())
 
    def writelines(self,list):
        map(self.write,list)


class AsyncIO(QObject): # {{{1
    """AsyncIO(self,parent=None)

    A class that implements asynchronous reading using the Qt event loop.
    parent is the optional parent QObject.  The lineReady signal is emitted
    when a complete line has been read.  The gotEOF signal is emitted if EOF
    was read.

    """
    def __init__(self,parent=None):
        QObject.__init__(self,parent)

        # There is no connection yet.
        self.disconnect()

    def disconnect(self):
        """
        Public method to disconnect any current connection.
        """
        self.readsn = None
        self.readfd = None

    def setDescriptors(self,rfd):
        """
        Public method called to set the descriptors for the connection.
        """
        self.rbuf = ''
        self.readfd = rfd

    def setNotifiers(self):
        """
        Protected method called to set up or take down the socket notifier for
        the Qt event loop.
        """
        if self.readsn is None:
            self.readsn = QSocketNotifier(self.readfd.fileno(),QSocketNotifier.Read)
            self.readsn.connect(self.readsn,SIGNAL('activated(int)'),self.readReady)
        else:
            self.readsn = None

    def readReady(self,fd):
        """
        Protected method called when there is data ready to be read.
        """
        # There seems to be a problem under Windows that the QSocketNotifier
        # says that there is data from the server, but the Python socket module
        # complains that the recv() on the socket would block and raises an
        # exception.  However, everything goes on to work as expected so we
        # just catch the exception and ignore it.
        try:
            got = self.readfd.readline()
        except:
            return

        if len(got) == 0:
            self.emit(PYSIGNAL('gotEOF'),())
            return

        self.rbuf = self.rbuf + got

        # Emit a signal for each complete line.
        eol = self.rbuf.find('\n')

        while eol >= 0:
            s = self.rbuf[:eol + 1]
            self.rbuf = self.rbuf[eol + 1:]
            self.emit(PYSIGNAL('lineReady'),(s,))
            eol = self.rbuf.find('\n')

# }}}1

class DebugClient(AsyncIO,bdb.Bdb):
    """DebugClient(self)

    Provide access to the Python interpeter from a debugger running in another
    process whether or not the Qt event loop is running.

    The protocol between the debugger and the client assumes that there will be
    a single source of debugger commands and a single source of Python
    statements.  Commands and statements are always exactly one line and may be
    interspersed.

    The client may send any other lines at any other time which should be
    interpreted as program output.

    If the debugger closes the session there is no response from the client.
    The client may close the session at any time as a result of the script
    being debugged closing or crashing.

    """
    def __init__(self):
        AsyncIO.__init__(self)
        bdb.Bdb.__init__(self)

        # The context to run the debugged program in.
        self.context = {'__name__' : '__main__'}

        # Make sure we have prompts.
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "

        # The list of complete lines to execute.
        self.buffer = ''

        self.connect(self,PYSIGNAL('lineReady'),self.handleLine)
        self.connect(self,PYSIGNAL('gotEOF'),self.sessionClose)

        self.clientReset()

    def clientReset(self):
        """
        Private method to reset everything for a new debugging session.
        """
        self.eventLoopDepth = 0
        self.inRawMode = 0
        self.fncache = {}
        self.veryFirstFrame = None
        self.firstAppFrame = None
        self.currentFrame = None
        self.debugFiles = []

        self.clear_all_breaks()
        self.reset()

    def preEventLoopHook(self):
        """
        Public method to handle the hook called when the Qt event loop is about
        to be entered.
        """
        if self.eventLoopDepth == 0:
            self.setNotifiers()
            self.flushResponse()

        self.eventLoopDepth = self.eventLoopDepth + 1

    def postEventLoopHook(self):
        """
        Public method to handle the hook called when the Qt event loop has
        terminated.
        """
        self.eventLoopDepth = self.eventLoopDepth - 1

        if self.eventLoopDepth == 0:
            self.setNotifiers()

    def raw_input(self,prompt):
        """
        Public method to implement raw_input() using the event loop.
        """
        self.write(prompt)
        self.inRawMode = 1
        self.eventLoop()
        return self.rawLine

    def sessionClose(self):
        """
        Private method to close the session with the debugger and terminate.
        """
        try:
            self.set_quit()
        except:
            pass

        self.disconnect()
        sys.exit()

    def handleLine(self,line):
        """
        Private method to handle a complete line by trying to execute the lines
        accumulated so far.
        """
        # Remove any newline (there should be one).
        if line[-1] == '\n':
                line = line[:-1]

        # If we are handling raw mode input then reset the mode and break out
        # of the current event loop.
        if self.inRawMode:
            self.inRawMode = 0
            self.rawLine = line
            self.eventExit = 1
            return

        eoc = line.find('<')

        if eoc >= 0 and line[0] == '>':
            # Get the command part and any argument.
            cmd = line[:eoc + 1]
            arg = line[eoc + 1:]

            if cmd == RequestLoad:
                self.clientReset()

                self.debugFiles, argl = eval(arg)
                os.chdir(argl[0])
                sys.argv = argl[1:]
                sys.path[0] = os.path.dirname(sys.argv[0])
                self.running = sys.argv[0]

                # This will eventually enter a local event loop.  The trick
                # with repr() deals with Windows path separators.
                self.run('execfile(%s)' % (repr(self.running)),self.context)
                return

            if cmd == RequestVariables:
                frm, scope, keys, filters = eval(arg)
                self.dumpDict(int(frm),int(scope),keys,filters)
                return

            if cmd == RequestStepOver:
                self.set_next(self.currentFrame)
                self.eventExit = 1
                return
                
            if cmd == RequestStep:
                self.currentFrame = None
                self.set_step()
                self.eventExit = 1
                return

            if cmd == RequestContinue:
                self.currentFrame = None
                self.set_continue()
                self.eventExit = 1
                return

            if cmd == RequestBreak:
                fn, line, set = eval(arg)
                line = int(line)
                set = int(set)

                if set:
                    self.set_break(fn,line)
                else:
                    self.clear_break(fn,line)

                return

        if self.buffer:
            self.buffer = self.buffer + '\n' + line
        else:
            self.buffer = line

        self.pendingResponse = sys.ps1

        try:
            code = codeop.compile_command(self.buffer,sys.stdin.name)
        except:
            # Report the exception to the user.
            sys.last_type, sys.last_value, sys.last_traceback = sys.exc_info()
            map(self.write,traceback.format_exception_only(sys.last_type,sys.last_value))
            self.buffer = ''
        else:
            if code is None:
                self.pendingResponse = sys.ps2
            else:
                self.buffer = ''

                try:
                    exec code in self.context
                except SystemExit, status:
                    self.progTerminated(status)
                except:
                    # Report the traceback to the user.
                    try:
                        type, value, tb = sys.exc_info()
                        sys.last_type = type
                        sys.last_value = value
                        sys.last_traceback = tb
                        tblist = traceback.extract_tb(tb)
                        del tblist[:1]
                        list = traceback.format_list(tblist)
                        if list:
                            list.insert(0,'Traceback (most recent call last):\n')
                            list[len(list):] = traceback.format_exception_only(type,value)
                    finally:
                        tblist = tb = None

                    map(self.write,list)

        self.flushResponse()

    def write(self,s):
        """
        Private method to print and flush some text.
        """
        sys.stdout.write(s)
        sys.stdout.flush()

    def flushResponse(self):
        """
        Private method to make sure any pending response is sent to the IDE.
        """
        if self.pendingResponse is not None:
            self.write(self.pendingResponse)
            self.pendingResponse = None

    def interact(self):
        """
        Interact with whatever is on stdin and stdout - usually the IDE.  This
        method will never terminate.
        """
        # Set the descriptors now and the notifiers when the Qt event loop is
        # entered for the first time.
        global DebugClientInstance

        self.setDescriptors(sys.stdin)
        DebugClientInstance = self

        # Tell the server what sys.path is set to.
        self.write('%s%s\n' % (ResponseSysPath,sys.path))

        # Announce ourselves.
        self.write('Python %s on %s\n' % (sys.version,sys.platform))
        self.write('Type "copyright", "credits" or "license" for more information.\n')
        self.write(sys.ps1)

        # At this point the Qt event loop isn't running, so simulate it.
        self.eventLoop()

    def eventLoop(self):
        self.eventExit = None

        while self.eventExit is None:
            if self.eventLoopDepth > 0:
                qApp.processOneEvent()
            else:
                rrdy, wrdy, xrdy = select.select([sys.stdin],[],[])

                if sys.stdin in rrdy:
                    self.readReady(sys.stdin.fileno())

        self.eventExit = None

    def connectDebugger(self,port):
        """
        Establish a session with the debugger and connect it to stdin.  We
        can't just use the existing stdin because Windows can't select on
        pipes, only sockets.
        """
        sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
        sock.connect(('127.0.0.1',port))
        sock.setblocking(0)

        sys.stdin = AsyncFile(sock,sys.stdin.mode,sys.stdin.name)

    def user_line(self,frame):
        """
        Re-implemented to handle the program about to execute a particular
        line.
        """
        # Ignore the very first frame.
        if frame is self.veryFirstFrame:
            return

        # Create the back trace.
        tlist = []

        f = frame

        classname = ''

        # receive classname if it is possible
        from types import InstanceType, MethodType
        code = frame.f_code
        name = code.co_name
        if code.co_argcount:
            arg0 = frame.f_locals[code.co_varnames[0]]
            if type(arg0) is InstanceType:
                attr = getattr(arg0, name, None)
                if attr and type(attr) is MethodType and attr.im_func.func_code is code:
                    classname = arg0.__class__.__name__

        while f is not None:
            tlist.append(self.absPath(f.f_code.co_filename))
            tlist.append(f.f_lineno)

            func = f.f_code.co_name

            if func == '?':
                func = ''

            if func != '' and classname != '':
                func = classname+'.'+func

            tlist.append(func)

            if f is self.firstAppFrame:
                f = None
            else:
                f = f.f_back

        self.currentFrame = frame

        self.write('%s%s\n' % (ResponseLine, str(tlist)))
        self.eventLoop()

    def user_exception(self,frame,(exctype, excval, exctb)):
        if exctype == SystemExit:
            self.progTerminated(excval)
        elif exctype is SyntaxError:
            try:
                msg, (fn, linenr, offset, text) = excval
            except:
                slist = []
            else:
                slist = [fn, linenr, offset, msg]

            self.write('%s%s\n' % (ResponseSyntax, str(slist)))

            self.eventLoop()
        else:
            if type(exctype) == types.ClassType:
                exctype = exctype.__name__

            if excval is None:
                excval = ''

            elist = [str(exctype), str(excval)]

            tblist = traceback.extract_tb(exctb)
            del tblist[:1]

            for tb in tblist:
                (fn, linenr, func, text) = tb

                if func == '?':
                    func = ''
    
                elist.append(fn)
                elist.append(linenr)
                elist.append(func)

            self.write('%s%s\n' % (ResponseException, str(elist)))

            self.eventLoop()

    def user_return(self,frame,retval):
        # The program has finished if we have just left the first frame.
        if frame == self.veryFirstFrame or frame == self.firstAppFrame:
            self.progTerminated(retval)

    def stop_here(self,frame):
        """
        Re-implemented to turn off tracing for certain files.
        """
        # Remember the very first frame.
        if self.veryFirstFrame is None:
            self.veryFirstFrame = frame

        # We always stop in the very first frame to allow things like syntax
        # errors to be detected.
        if self.veryFirstFrame is frame:
            return 1

        # Remember the first application frame.
        if self.firstAppFrame is None:
            self.firstAppFrame = frame

        # Always ignore line 0.
        if frame.f_lineno == 0:
            return 0

        fn = frame.f_code.co_filename

        # Check that we are interested in the file.
        fn = self.absPath(fn)

        if not fn in self.debugFiles:
            return 0

        return bdb.Bdb.stop_here(self,frame)

    def dumpDict(self,frmnr,scope,key,filters):
        """
        Private method to return dictionary from a particular frame.
        """
        # Find the frame.
        f = self.currentFrame

        while f is not None and frmnr > 0:
            f = f.f_back
            frmnr = frmnr - 1

        if f is None:
            return

        if scope:
            dict = f.f_globals
        else:
            dict = f.f_locals

            if f.f_globals is f.f_locals:
                scope = 2

        vlist = [scope, key]
        dtype = None

        if scope != 2:
            # Find the right dictionary.
            if key:
                for k in key.split('.'):
                    # The key component might be quoted.
                    try:
                        k = eval(k)
                    except:
                        pass

                    try:
                        dict = dict[k]
                    except:
                        dict = None

                    dtype = type(dict)

                    if dtype is types.InstanceType:
                        dict = dict.__dict__
                    elif dtype is not types.DictType:
                        return

            # Convert the filters string to a list.
            flist = []

            if filters:
                filters = filters.split('.')

                for f in filters:
                    flist.append(eval(f))

            # Return each key/attribute in the dictionary.
            klist = dict.keys()
            klist.sort()

            for k in klist:
                # Filter 0 is for hidden attributes.
                if 0 in flist and str(k)[:2] == '__':
                    continue

                v = dict[k]
                vtype = self.getType(v)

                if vtype in flist:
                    continue

                # Replace an unknown type by its name.
                if vtype == 26:
                    vtype = str(type(v))[7:-2]

                # See the note above about the string keys of the frame
                # dictionary.  Here we make sure keys of ordinary dictionaries
                # are quoted, but other dictionaries are not.
                if dtype is types.DictType:
                    vlist.append(repr(k))
                else:
                    vlist.append(k)

                vlist.append(repr(v))
                vlist.append(vtype)

        self.write('%s%s\n' % (ResponseVariables, str(vlist)))

    def absPath(self,fn):
        """
        Private method to convert a filename to an absolute name using sys.path
        as a set of possible prefixes. The name stays relative if a file could
        not be found.
        """
        if os.path.isabs(fn):
            return fn

        # Check the cache.
        if self.fncache.has_key(fn):
            return self.fncache[fn]

        # Search sys.path.
        for p in sys.path:
            afn = os.path.normpath(os.path.join(p,fn))

            if os.path.exists(afn):
                self.fncache[fn] = afn
                return afn

        # Nothing found.
        return fn

    def progTerminated(self,status):
        """
        Private method to tell the debugger that the program has terminated.
        """
        if status is None:
            status = 0

        self.set_quit()
        self.write('%s%s\n' % (ResponseExit,status))

    def getType(self,v):
        """
        Private method to convert the type of an object to a numerical value.
        Don't change these values without making corresponding changes to the
        C++ code.
        """
        typ = type(v)

        if typ is types.NoneType:
            return 1

        if typ is types.TypeType:
            return 2

        if typ is types.IntType:
            return 3

        if typ is types.LongType:
            return 4

        if typ is types.FloatType:
            return 5

        if typ is types.ComplexType:
            return 6

        if typ is types.StringType:
            return 7

        if typ is types.UnicodeType:
            return 8

        if typ is types.TupleType:
            return 9

        if typ is types.ListType:
            return 10

        if typ is types.DictType:
            return 11

        if typ is types.FunctionType:
            return 12

        if typ is types.CodeType:
            return 13

        if typ is types.ClassType:
            return 14

        if typ is types.InstanceType:
            return 15

        if typ is types.MethodType:
            return 16

        if typ is types.BuiltinFunctionType:
            return 17

        if typ is types.ModuleType:
            return 18

        if typ is types.FileType:
            return 19

        if typ is types.XRangeType:
            return 20

        if typ is types.SliceType:
            return 21

        if typ is types.EllipsisType:
            return 22

        if typ is types.TracebackType:
            return 23

        if typ is types.FrameType:
            return 24

        if typ is types.BufferType:
            return 25

        # Some other type.
        return 26


# We are normally called by the debugger to execute directly.

if __name__ == '__main__':
    # Get the port to connect to the debugger to.
    try:
        port = int(sys.argv[1])
    except:
        port = -1

    sys.argv = ['']
    sys.path[0] = ''
    debugClient = DebugClient()

    # Connect to the debugger if we have a port.  Under UNIX omitting the port
    # means we have an interactive interpreter which is useful for testing.  It
    # doesn't work under Windows because Windows is cack.
    if port >= 0:
        debugClient.connectDebugger(port)

    debugClient.interact()
