#!/usr/bin/env python

import sys
from qt import *

TRUE  = 1
FALSE = 0

class Canvas(QWidget):
    def __init__(self,parent=None,name=None):
        QWidget.__init__(self,parent,name,Qt.WStaticContents)
        self.pen = QPen(Qt.red,3)
        self.no_writing = FALSE
        self.polyline = QPointArray(3)
        self.mousePressed = FALSE
        self.buffer = QPixmap(self.width(),self.height())
        self.buffer.fill(self.colorGroup().base())
        self.setBackgroundMode(QWidget.PaletteBase)
        self.setCursor(Qt.crossCursor)

    def setPenColor(self,c):
        self.pen.setColor(c)

    def setPenWidth(self,w):
        self.pen.setWidth(w)

    def penColor(self):
        return self.pen.color()

    def penWidth(self):
        return self.pen.width()

    def save(self,filename,format):
        if not self.no_writing:
            self.buffer.save(filename,format.upper())

    def clearScreen(self):
        self.buffer.fill(self.colorGroup().base())
        self.repaint(FALSE)

    def mousePressEvent(self,e):
        self.mousePressed = TRUE
        for i in range(0,3):
            self.polyline.setPoint(i,e.pos().x(),e.pos().y())

    def mouseReleaseEvent(self,e):
        self.mousePressed = FALSE

    def mouseMoveEvent(self,e):
        if self.mousePressed:
            painter = QPainter()
            painter.begin(self.buffer)
            painter.setPen(self.pen)
            for i in range(-2,0):
                x = self.polyline.point(-i-1)[0]
                y = self.polyline.point(-i-1)[1]
                self.polyline.setPoint(-i,x,y)
            self.polyline.setPoint(0,e.pos().x(),e.pos().y())
            painter.drawPolyline(self.polyline)
            painter.end()

            r = self.polyline.boundingRect()
            r = r.normalize()
            r.setLeft(r.left() - self.penWidth())
            r.setTop(r.top() - self.penWidth())
            r.setRight(r.right() + self.penWidth())
            r.setBottom(r.bottom() + self.penWidth())
    
            bitBlt(self,r.x(),r.y(),self.buffer,r.x(),r.y(),r.width(),r.height())

    def resizeEvent(self,e):
        QWidget.resizeEvent(self,e)

        if self.width() > self.buffer.width():
            w = self.width()
        else:
            w = self.buffer.width()

        if self.height() > self.buffer.height():
            h = self.height()
        else:
            h = self.buffer.height()

        tmp = QPixmap(self.buffer)
        self.buffer.resize(w,h)
        self.buffer.fill(self.colorGroup().base())

        bitBlt(self.buffer,0,0,tmp,0,0,tmp.width(),tmp.height())

    def paintEvent(self,e):
        QWidget.paintEvent(self,e)
    	rects = []
        rects.append(e.rect())
    	for i in rects:
            r = i
            bitBlt(self,r.x(),r.y(),self.buffer,r.x(),r.y(),r.width(),r.height())

class Scribble(QMainWindow):
    def __init__(self):
        QMainWindow.__init__(self)

        self.canvas = Canvas(self)
        self.setCentralWidget(self.canvas)

        tools = QToolBar(self)

        self.bSave = QToolButton(QIconSet(),'Save as','Save as',self.slotSave,tools)
        self.bSave.setText('Save as...')

        tools.addSeparator()

        self.bPColor = QToolButton(QIconSet(),'Choose Pen Color','Choose Pen Color',self.slotColor,tools)
        self.bPColor.setText('Choose Pen Color...')

        tools.addSeparator()

        self.bPWidth = QSpinBox(1,20,1,tools)
        QToolTip.add(self.bPWidth,'Choose Pen Width')
        self.connect(self.bPWidth,SIGNAL('valueChanged(int)'),self.slotWidth)
        self.bPWidth.setValue(3)

        self.bClear = QToolButton(QIconSet(),'Clear Screen','Clear Screen',self.slotClear,tools)
        self.bClear.setText('Clear Screen')

    def slotSave(self):
        menu = QPopupMenu(None)
        formats = {}
        for i in QImage.outputFormats():
            formats[menu.insertItem(str(i).upper()+'...')] = str(i)
        menu.setMouseTracking(TRUE)
        _id = menu.exec_loop(self.bSave.mapToGlobal(QPoint(0,self.bSave.height()+1)))
        if _id != -1:
            format = str(formats[_id])
            filename = QFileDialog.getSaveFileName('','*.'+format.lower(),self)
            if not filename.isEmpty():
                
                self.canvas.save(filename,format)

    def slotColor(self):
        c = QColorDialog.getColor(self.canvas.penColor(),self)
        if c.isValid():
            self.canvas.setPenColor(c)

    def slotWidth(self,int):
        self.canvas.setPenWidth(int)

    def slotClear(self):
        self.canvas.clearScreen()

## main program
a = QApplication(sys.argv)

scribble = Scribble()
scribble.resize(500,350)
scribble.setCaption("PyQT Example - Scribble")
a.setMainWidget(scribble)
scribble.show()

a.exec_loop()